#!/usr/bin/env python
from tulip.tasks import coroutine
from tulip.locks import EventWaiter

class ROBuffer:
    def __init__(self, inital=b""):
        self._buffer = buffer
        self._future = None
        self._finialized = EventWaiter()
        self.finialized = False
        
    @coroutine
    def read(self, length=None):
        if length and length < len(self._buffer):
            buf = self._buffer[:length]
            self._buffer = self._buffer[length:]
            return buf
        elif length is None:
            yield from self._finalized.wait()
            buf = self._buffer
            self._buffer = b""
            return buf
        else:
            # waiting on more data to read
            while True:
                yield from self._append_event.wait()
                if length < len(self.buffer):
                    buf = self._buffer[:length]
                    self._buffer = self._buffer[length:]
                    return buf
                
        
    def append_buffer(self, data):
        self._buffer += buffer
        self._future.set_result(buffer)
        self._append_event.set()
    
    def finalize(self):
        """No more data to be read, buffer is 'finalized'"""
        self.finalized = True
        self._finalized.set()
        self._future.set_result()

class OutputStream:
    """A Buffer for outgoing data transmission on a file like object.
    
    this implementation can be 'paused', which when activated will buffer
    all writes internnaly and return instantly. when the buffer is 'unpaused'
    all pending writes will be commited in one go.
    
    this is mainly intended to prevent writing of an HTTP reply body before
    the headers have been written (ie start_response will write the headers 
    to the stream then unpause the buffer) but may also be used for coalescing
    writes together and sending them as one update
    """
    def __init__(self, stream):
        self._stream = stream
        
        self.is_paused = True
        self._buffer
    
    def sendfile(self, filename):
        pass
    
    def write(self, data):
        if self.is_paused:
            self._buffer += data
        
    def writelines(self, lines):
        if self.is_paused:
            self._buffer = b"".join([self._buffer] + lines)
        else:
            # selector based implementation, for a proactor
            # based implementation look at b"".join(lines)
            # and submitting that as one unit or using a
            # scatter/gather write mechanism
            for line in lines:
                self.write(line)
        
    def pause(self):
        self.is_paused = True
        
    def unpause(self):
        self.is_paused = False
        self._flush()
    
    def _flush(self):
        pass
        



#!/usr/bin/env python
from tulip.protocols import Protocol
from tulip.tasks import task, sleep
from tulip import get_event_loop
from .process import Process
from io import BytesIO

class HTTPProtocolWarning(Exception): pass
class HTTPProtocolError(Exception): pass

class BaseHTTPProtocol(Protocol):
    def connection_made(self, transport):
        self.transport = transport
#        self.body = body = None #Buffer()
#        self.handle_request('GET', '/', 'HTTP/1.1', {}, body)

    def data_recived(self, data):
#        self.body.append_buffer(data)
        pass
                
    def eof_recived(self):
        pass
        
    def connection_lost(self, exc):
        pass
    
class HTTPServerProtocol(BaseHTTPProtocol, Process):
    def __init__(self, app):
        self.app = app
        self.headers = []
        self.body = BytesIO()
        self.state = 'parse protocol'
        self.buffer = b""
        
    def data_received(self, data):
        try:
            if self.state == 'parse protocol':
                # we check for a newline in data rather self.buffer
                # as it reduces the ammount of bytes we have to search
                # also means each byte is only checked once, where ase
                # searching self.buffer means the leading bytes could
                # be checked multiple times
                if b'\n' in data:
                    line, sep, new_buffer = data.partition(b'\n')
                    line = self.buffer + line.strip()
                    self.method, self.path, self.protocol = self.parse_protocol(line)
    
                    data = new_buffer
                    self.buffer = b""
                    
                    self.state = 'parse headers'
                else:
                    self.buffer += data
                    
            if self.state == 'parse headers':
                if b'\r\n\r\n' in data:
                    block, sep, new_buffer = data.partition(b'\r\n\r\n')
                    block = self.buffer + block
                    self.headers = self.parse_headers(block)
                    
                    self.body.seek(0)
                    self.body.write(new_buffer)
                    
                    self.state = 'body'

                    loop = get_event_loop()
                    loop.call_soon(self.app, self.transport, 
                                             self.method, 
                                             self.path, 
                                             self.protocol, 
                                             self.headers, 
                                             self.body)
                elif b'\n\n' in data:
                    block, sep, new_buffer = data.partition(b'\n\n')
                    block = self.buffer + block
                    self.headers = self.parse_headers(block)
                    
                    self.body.seek(0)
                    self.body.write(new_buffer)
                    
                    self.state = 'body'
                    
                    loop = get_event_loop()
                    loop.call_soon(self.app, self.transport, 
                                             self.method, 
                                             self.path, 
                                             self.protocol, 
                                             self.headers, 
                                             self.body)
                else:
                    self.buffer += data
                    
            if self.state == 'body':
                self.body.write(data)

        except HTTPProtocolWarning:
            # TODO: Log a warning
            pass
        except HTTPProtocolError:
            # Bad unrecoverable things have happened, commit Seppuku
            self.transport.close()
            
    @staticmethod
    def parse_protocol(line):
        components = line.split()
        if len(components) != 3:
            raise VeryBadError('malformed request, Protocol line: ' + line)
        
        return components
        
    @staticmethod
    def parse_headers(block):
        # can we cut down the ammount of allocations here?
        lines = block.split(b'\n')
        # [::2] will skip the middle element (sep) that we dont want
        headers = (line.partition(b':')[::2] for line in lines)
        headers = [(x.strip(), y.strip()) for x,y in headers]
        
        return headers

#    @task
#    def handle_request(self):
#        self.transport.write(b'200 OK\r\ncontent-length: 4\r\n\r\ntest')
#        self.transport.close()
#        self.exit()


class FCGIProtocol(BaseHTTPProtocol, Process):
    def __init__(self, app):
        self.app = app
        
    @task
    def handle_request(self):
        self.transport.write(b'200 OK\r\ncontent-length: 4\r\n\r\ntest')
        self.transport.close()
        self.exit()




#!/usr/bin/env python
"""TWIG: An collection of helper functions for implementing the tulip webservices interface gateway
"""
__author__ = "Da_Blitz"
__email__ = "code@pocketnix.org"
__url__ = "http://code.pocketnix.org/archangel"
__version__ = "0.3"
__license__ = "ISC"

from urllib.parse import quote_from_bytes as urlencode
import io

HTTP_REPLY_VERSION = b'HTTP/1.1'
LF = b"\r\n"

def split_version(raw_version):
    """Given a string representing a version as specified in RF3875, transform it into a tuple
    
    eg:
    >>> split_version(b'HTTP/1.1')
    (b'HTTP', (1, 1))
    
    This function does not have the limitation of the 2 version numbers being single digits
    
    :param bytes raw_version: A Byte string representing the version to be parsed
    :rtype tuple: The parsed version contisting of a tuple containg the protocol followed by a 2
                  tuple containg 2 integers representing the version number
    
    May raise a ValueError if the supplied version numbers are not integers
    """
    protocol, version = raw_version.split(b"/", 1)
    major, minor = version.split(b".")
    
    major, minor = int(major), int(minor)
    
    return (protocol, (major, minor))


class Request:
    def __init__(self, method, headers={}, body=None, script_name=b"", path_info=b"", query=b"", 
                 content_type=None, content_length=None, envrion={},
                 server_name=b"", server_addr=b"", server_port=80, 
                 server_protocol=(b"HTTP", (0, 9)), server_type=None,
                 remote_addr=b"", remote_host=b"", remote_port=0, remote_user=None,
                 ssl_protocol=None, auth_type=None, gateway_interface=(b"TWIG", (0, 1))):
        # method: All uppercase byte string, eg: [HEAD GET PUT POST DELETE PATCH]
        self.method = method
        # headers: dict
        self.headers = headers
        # body: a filelike object compatible with tulip (future from .read()) that is non-seekable
        self.body = body
        # script_name: byte string
        self.script_name = script_name
        # path_info: byte string
        self.path_info = path_info
        # query: byte string
        self.query = query
        # content_type: byte string
        self.content_type = content_type
        # content_length: None or int, None for no body methods such as GET, int otherwise
        #                 -1: No length specified (possible chuncked body encoding
        #                  0: Body has no content
        #                >=1: The length in bytes of the request body
        # TODO: Can we eliminate teh None state and make it '0', perhaps make -1 'None'
        if content_length:
            content_length = int(content_length)
        self.content_length = content_length
        # environ: dict
        self.envrion = environ
        
        # server_name: byte string
        self.server_name = server_name
        # server_addr: byte_string
        self.server_addr = server_addr
        # server_port: int
        self.server_port = int(server_port)
        # server_protocol: protocol version as specified by split_version()
        if isinstance(server_protocol, bytes):
            server_protocol = split_version(server_protocol)
        self.server_protocol = server_protocol
        # server_version: server version as specified by split_version()
        if isinstance(server_type, bytes):
            server_type = split_version(server_type)
        self.server_type = server_type
    
        # remote_addr: byte_string
        self.remote_addr = remote_addr
        # remote_host: byte_string
        self.remote_host = remote_host
        # remote_port: int
        self.remote_port = int(remote_port)
        # remote_user: byte_string
        self.remote_user = remote_user
        
        # ssl_protocol: ssl version as specified by split_version()
        if isinstance(ssl_protocol, bytes):
            ssl_protocol = split_vertion(ssl_protocol)
        self.ssl_protocol = ssl_protocol
        # auth_type: byte string, uppercase, eg: b'DIGEST' or b'SPENGO'
        self.auth_type = auth_type.upper()
        # gateway_interface: gateway interface version as specified by split_version()
        if isinstance(gateway_interface, bytes):
            gateway_interface = split_version(gateway_interface)
        self.gateway_interface = gateway_interface
    
    @property
    def path_translated(self):
        """A url encoded version of :py:attr:`.path_info`
        
        :rtype bytes: The url encoded path_info
        """
        return urlencode(self.path_info)
    
    @property
    def secure(self):
        """Returns True if using an encrypted connection to the client
        
        :rtype bool: is client connecting using an encrypted connection?
        """
        if self.ssl_protocol:
            return self.ssl_protocol[0] in ['TLS', 'SSL']
        return false
    
    https = secure

    @property
    def authenticated(self):
        """Resturns True if the remote user has authenticated to the server
        
        :rtype bool: has the client authenticated?
        """
        return self.remote_user is not None

    def __getitem__(self, key):
        return self.headers[key]

    def __contains__(self, key):
        return key in self.headers

    def clone(self):
        """Return a copy of ourselves that can be modified, then supplied to the next app in a middleware pipe
        
        :rtype Response: A clone of the current response object
        """
        obj = self.__new__(self.__class__)
        obj.__dict__ = self.__dict__.copy()
        
        return obj
        

class Response:
    def __init__(self, transport):
        self.transtport = transport
        self.clear()
        self._paused = True
        
    def add_header(self, key, val):
        self.headers.append((key, val))
        
    def start_response(self, status, headers, *, version=HTTP_REPLY_VERSION):
        """No data is sent until this function is called. all data will be buffered"""
        # Lots of small writes comming up, so collapse them into one larger write
        self.pause()
        self.write(version, status + LF)
        headers = self.headers + headers
        self.writelines(k + b": " + v + LF for k, v in headers)
        self.write(LF)
        self.unpause()
        
        # Stop buffering writes
        self._unpause()
        self._flush()
        
    def close(self):
        """Flush all remaining data and close the connection
        
        Applications should not normmaly need to call this as connections should
        be closed by your framework when the function returns
        """
        self.transport.close()
    
    def abort():
        """Close the connection to the other end immediately, don't flush data"""
        self.transport.abort()
    
    def pause(self):
        """Simmilar to TCP_CORK"""
        self.transport.pause()
        
    def resume(self):
        self.transport.resume()
        
    def _unpause(self):
        """Called by start_response to indicate that the headers have been written and that data can now
        be submitted for writing to the socket
        
        Note: this is diffrent to the pause and unpause methods that are useful for 'bunching' up lots
        of small write requests into one larger one to reduce overhead
        """
        self._paused = False
    
    def _flush(self):
        for obj in self._buffer:
            if isinstance(obj, bytes):
                self.transport.write(obj)
            elif isisntance(obj, io.IOBase):
                # _pause should be false by now, so sendfile whould do what we want
                self.sendfile(obj)
            else:
                raise ValueError("Unknown datatype in buffer (did you encode all text?): {}".format(obj))
        self._buffer = []
    
    def write(self, data):
        if self._paused:
            self._buffer.append(data)
        else:
            self.transport.write(data)
            
    def writelines(self, lines):
        if self._paused:
            self._buffer += lines
        else:
            self.transport.writelines(lines)
            
    def sendfile(self, file, chunk_size=io.DEFAULT_BUFFER_SIZE):
        if not isintance(file, file):
            file = open(file, 'rb')
        
        if self._paused:
            self._buffer.append(file)
        else:
            buf = file.read(chunk_size)
            while len(buf) != 0:
                self.transport.write(buf)
                buf = file.read(chunk_size)
                
        
    def clear(self):
        """clear the contents of the buffer and headers, for apps that do everything rather than rely on midleware
        allows them to zero the buffers contents before sending an error page if an exception was raised during
        rendering of the page
        """
        self.headers = []
        self._buffer = []



#!/usr/bin/env python

def make_environ():
    uri_parts = urlsplit(info.uri)
    url_scheme = 'https' if self.is_ssl else 'http'
    
    environ = {
        'wsgi.input': payload,
        'wsgi.errors': sys.stderr,
        'wsgi.version': (1, 0),
        'wsgi.async': True,
        'wsgi.multithread': False,
        'wsgi.multiprocess': False,
        'wsgi.run_once': False,
        'wsgi.file_wrapper': FileWrapper,
        'wsgi.url_scheme': url_scheme,
        'SERVER_SOFTWARE': tulip.http.HttpMessage.SERVER_SOFTWARE,
        'REQUEST_METHOD': info.method,
        'QUERY_STRING': uri_parts.query or '',
        'RAW_URI': info.uri,
        'SERVER_PROTOCOL': 'HTTP/%s.%s' % info.version
    }



#!/usr/bin/env python
"""Archangel: fast lazy webserver based off of TULIP/yield from pep-3156
"""
__author__ = "Da_Blitz"
__email__ = "code@pocketnix.org"
__url__ = "http://code.pocketnix.org/archangel"
__version__ = "0.3"
__license__ = "ISC"




#!/usr/bin/env python
from .protocols import HTTPServerProtocol
from .process import ProcessManager
from tulip.tasks import task
import tulip
import sys


def start_server(app, addr, manager=None, transport=HTTPServerProtocol, **kwargs):
    """Returns the listening socket so we can do things like add more flags to it"""
    loop = tulip.get_event_loop()

    if manager:
        servlet = lambda: manager.register(transport(app))
    else:
        servlet = lambda: transport(app)
        
    server = loop.start_serving(servlet, *addr, **kwargs)
    loop.run_until_complete(server)

def add_shutdown_handler(manager=None):
    import signal

    def sigint_handler():
        print("Recived SIGINT, Exiting gracefully")
        if manager:
            manager.shutdown()
        sys.exit(0)
        
    loop = tulip.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, sigint_handler)


@task
def dummy_app(transport, method, path, protocol, headers, body):
    transport.write(b'HTTP/1.1 200 OK\r\ncontent-type: text/plain\r\n\r\n')
    method = method.decode('utf-8')
    path = path.decode('utf-8')
    protocol = protocol.decode('utf-8')
    transport.write("Protocol: {} {} {}:\n".format(method, path, protocol).encode('utf-8'))
    transport.write("Headers:\n".encode('utf-8'))
    for key, val in headers:
        key = key.decode('utf-8')
        val = val.decode('utf-8')
        line = "{:<20}{}\n".format(key + ":", val)
        transport.write(line.encode('utf-8'))
    transport.close()

def tulip_server(argv=sys.argv[1:]):
    loop = tulip.get_event_loop()

    manager = ProcessManager()
    start_server(dummy_app, ('localhost', 8080), manager)
    sock = add_shutdown_handler(manager)

#    print('Listening on {}:{}'.format(*sock.getsockname()))
    loop.run_forever()
    
if __name__ == "__main__":
    tulip_server()



#!/usr/bin/env python
from weakref import WeakValueDictionary

class ProcessManager:
    def __init__(self):
        self.processes = WeakValueDictionary()
        self.pid = 0
        self._next_pid = self.pid + 1
    
    def new_pid(self):
        pid = self._next_pid
        self._next_pid += 1
        
        return pid
    
    def register(self, obj):
        pid = self.new_pid()
        obj._archangel_pid = pid
        obj._archangel_manager = self
        self.processes[pid] = obj

        return obj
    
    def deregister(self, pid):
        del self.processes[pid]

    def shutdown(self):
        for process in self.processes.values():
            kill(process)

class Process:
    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        obj._at_exit = []
        # add a handler to signal the Process manager that we have exited
        obj.at_exit(lambda proc: proc._archangel_manager.deregister(proc._archangel_pid))
        
        return obj
        
    def _kill(self):
        try:
            self.kill()
        except:
            # ensure that we exit corectly even if an exception is raised
            self.exit()
            raise
        else:
            self.exit()
        
    def kill(self):
        """Function hook for implementations to catch a 'kill' signal and do some 
        cleanup work
        """
        pass
        
    def at_exit(self, func, **kwargs):
        """Register a function to be called when this process exits
        
        each function registered will be called with the process instance as the
        first argument
        """
        self._at_exit.append((func, kwargs))
        
    def exit(self):
        """Notify the ProcessManager that we are done or will finish up shortly"""
        for func, kwargs in self._at_exit:
            try:
                func(self, **kwargs)
            except Exception as err:
                # TODO: log this as an exception
                pass
            
def kill(proc):
    """Send the exit signal to a process"""
    proc._kill()
